#!/usr/bin/python
# -*- coding: utf-8 -*-
#
#       Copyright (C) 2007-2008 Simonenko Sergey <gforgx@gmail.com>
#
#       Redistribution and use in source and binary forms, with or without
#       modification, are permitted provided that the following conditions are
#       met:
#
#       * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#       * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following disclaimer
#       in the documentation and/or other materials provided with the
#       distribution.
#       * Neither the name of the project NoteFinder and author's name nor the names of 
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
#       THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#      "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#       LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#       A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#       OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#       SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#       LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#       DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#       THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#       (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#       OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import os
import ConfigParser
import xmlrpclib
from urllib import urlencode, urlopen
from icalendar import Calendar, Journal
from icalendar.prop import vText
from time import localtime
from poplib import POP3
from email import message_from_string
from email.header import decode_header
from email.iterators import typed_subpart_iterator
try:
    from feedparser import parse
    RSSParser = True
except ImportError:
    RSSParser = False

def today():
    """ Represents current date """
    return '%d-%d-%d' % localtime()[:3]

class Config(ConfigParser.ConfigParser):
    """ Configuration parser. Configuration file is a
    simple INI stored in ~/.config directory """
    def __init__(self):
        """ Initializes ConfigParser and reads
        existing configuration file into memory """
        ConfigParser.ConfigParser.__init__(self)
        self.optionxform = str
        self.file = os.path.expanduser("~/.config/notefinder/notefinder.ini")
        confDir = os.path.expanduser("~/.config")
        if not os.path.exists(confDir):
            os.mkdir(confDir)
        if not os.path.exists(os.path.join(confDir, "notefinder")):
            os.mkdir(os.path.join(confDir, "notefinder"))
        if os.path.exists(self.file):
            self.read(self.file)
    
    def addNotebook(self, name, backend, markup, settings):
        """ Sets specifed "name" notebook option to "backend" value
        into Notebooks section and to stringified pseudo-list
        "settings" value into "backend" section """
        if not self.has_section("Notebooks"):
            self.add_section("Notebooks")
        self.set("Notebooks", name, backend)
        if not self.has_section(backend):
            self.add_section(backend)
        self.set(backend, name, ";".join(settings))
        if not self.has_section("Markup"):
            self.add_section("Markup")
        self.set("Markup", name, markup)
        self.write(open(self.file, 'w'))
        
    def deleteNotebook(self, name):
        self.remove_option(self.getBackend(name), name)
        self.remove_option("Notebooks", name)
        self.write(open(self.file, 'w'))
    
    def setDefault(self, name):
        """ Sets "Notebook" option to "name" value
        in "General section """
        if not self.has_section("General"):
            self.add_section("General")
        self.set("General", "Notebook", name)
        self.write(open(self.file, 'w'))

    def getNotebook(self):
        """ Returns default notebook name """
        if self.has_option("General", "Notebook"):
            return self.get("General", "Notebook")

    def getNotebooks(self):
        """ Returns a list of all presented notebook """
        if self.has_section("Notebooks"):
            return self.options("Notebooks")

    def getBackend(self, name):
        """ Returns a string respresentation
        of backend for notebook """
        if self.has_option("Notebooks", name):
            return self.get("Notebooks", name)
 
# Initializing configuration parser
config = Config()

### BACKENDS ###

class FileSystem(object):
    """ Backend for storing entries on file system """
    
    desc = 'Backend for storing entries on file system'
    
    ### SETTINGS DESCRIPTION ###
    path = True
    url = False
    login = False
    passwd = False
    
    ### FEATURE SET ###
    Tag = True
    ReadOnly = False
    URL = True
    
    def __init__(self, notebook, path = None):
        """ Initializing everything.
        If "path" is specified backend will use it instead
        of configuration file entry"""
        self.name = "FileSystem"
        self.notebook = notebook

        # Configuration
        if path is None:
            path = config.get(self.name, self.notebook)
        
        # Creating root directory if not exists
        if not os.path.exists(path):
            os.mkdir(path)

        # Subdirectories
        self.path = os.path.join(path, 'date')
        self.tagPath = os.path.join(path, 'tag')

        # Creating subdirectories if not exist
        for i in (self.path, self.tagPath):
            if not os.path.exists(os.path.join(path, i)):
                os.makedirs(os.path.join(path, i))

    def getDates(self):
        """ Returns listing of dates root directory """
        return os.listdir(self.path)

    def getTags(self):
        """ Returns listing of tags root directory """
        return os.listdir(self.tagPath)

    def getURL(self, entry):
        return "file://" + self.getPath(entry)
        
    def getNotes(self):
        """ Walks through date directories
        and collects list of **all** entries """
        notes = []
        for i in self.getDates():
            notes += os.listdir(os.path.join(self.path, i))
        return notes

    def getNotesByDate(self, date):
        """ Returns contents of date directory """
        if self.dateExists(date):
            return os.listdir(os.path.join(self.path, date))
        else:
            return []

    def getNotesByTag(self, tag):
        """ Returns contents of tag directory """
        if self.tagExists(tag):
            return os.listdir(os.path.join(self.tagPath, tag))
        else:
            return []

    def getPath(self, entry):
        """ Walks through date directories
        and returns entry path """
        path = None
        for date in self.getDates():
            if entry in self.getNotesByDate(date):
                path = os.path.join(self.path, date, entry)
                # Breaks cycle after finding first matching date dir
                break
        return path

    def getText(self, entry):
       """ Opens file for reading and returns text """
       return "".join(open(self.getPath(entry)).readlines())

    def getNoteDate(self, entry):
        """ Returns date """
        return os.path.split(
            os.path.split(self.getPath(entry))[0])[1]

    def getNoteTags(self, entry):
        """ Returns tags """
        return [tag for tag in self.getTags() if entry in self.getNotesByTag(tag)]

    def dateExists(self, date):
        """ Checks if date is in listing of  dates root directory """
        return (date in self.getDates())

    def tagExists(self, tag):
        """ Checks if tag is in listing of tags root directory """
        return (tag in self.getTags())

    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())

    def createDate(self, date):
        """ Creates date directory inside dates root """
        if not self.dateExists(date):
            os.mkdir(os.path.join(self.path, date))

    def createTag(self, tag):
        """ Creates tag directory inside tags root """
        if not self.tagExists(tag):
            os.mkdir(os.path.join(self.tagPath, tag))

    def removeDate(self, date):
        """ Removes date directory if it's empty"""
        if self.dateExists(date) and len(self.getNotesByDate(date)) == 0:
            os.rmdir(os.path.join(self.path, date))

    def removeTag(self, tag):
        """ Removes tag directory and all symlinks inside """
        if self.tagExists(tag):
            for i in self.getNotesByTag(tag):
                os.remove(os.path.join(self.tagPath, tag, i))
            os.rmdir(os.path.join(self.tagPath, tag))

    def deleteNote(self, entry):
        """ Deletes entry file and all symlinks """
        if self.noteExists(entry):
            date = self.getNoteDate(entry)
            tags = self.getNoteTags(entry)
            os.remove(self.getPath(entry))
            self.removeDate(date)
            for tag in tags:
                os.remove(os.path.join(self.tagPath, tag, entry))
                # Removing tag if it has no any note
                if len(self.getNotesByTag(tag)) == 0:
                    self.removeTag(tag)

    def write(self, entry, text):
        """ Writes specified text to an entry file """
        if not self.noteExists(entry):
            date = today()
            self.createDate(date)
            open(os.path.join(self.path, date, entry), 'w').write(text)
        else:
            open(self.getPath(entry), 'w').write(text)

    def tag(self, entry, tags):
        """ Creates symlinks for entry in each tag's directory
        in tags root """
        src = self.getNoteTags(entry)
        for tag in tags:
            if tag not in src:
                self.createTag(tag)
                try:
                    # Creates non-absolute symbolic link to entry file
                    os.symlink(os.path.join("../../date",
                        self.getNoteDate(entry), entry),
                        os.path.join(self.tagPath, tag, entry))
                except AttributeError:
                    # Hack for filesystems without symbolic links support
                    open(os.path.join(self.tagPath, tag, entry), 'w')
        for tag in src:
            if tag not in tags:
                os.remove(os.path.join(self.tagPath, tag, entry))
                self.removeTag(tag)

class iCal(object):
    """ iCalendar backend """

    desc = 'iCalendar backend'
    
    ### SETTINGS DESCRIPTION ###
    path = True
    url = False
    login = False
    passwd = False

    ### FEATURE SET ###
    Tag = False
    ReadOnly = True
    URL = False

    def __init__(self, notebook):
        self.name = "iCal"
        self.notebook = notebook

        self.path = config.get(self.name, self.notebook)

        # Raising exception if path is wrong
        if not os.path.exists(self.path):
                raise NotebookError("iCalendar file doesn't exist")
        
        if os.path.isdir(self.path):
                raise NotebookError("Specified path isn't an iCalendar file")

    def getNotes(self):
        """ Returns listing of <mountpoint>/Notes """
        return [str(i["SUMMARY"]) for i in Calendar.from_string(open(self.path).read()).subcomponents]
    
    def getDates(self):
        """ Workaround """
        return [str(i["CREATED"].dt).split(" ")[0] for i in Calendar.from_string(open(self.path).read()).subcomponents]

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        pass

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today():
            return self.getNotes()
        else:
            return []
    
    def getNotesByTag(self, tag):
        return []
    
    def getText(self, entry):
        for i in Calendar.from_string(open(self.path).read()).subcomponents:
            if str(i["SUMMARY"]) == entry:
                text = vText.from_ical(i["DESCRIPTION"].ical()).encode("utf8")
                break
        return text
    
    def getNoteDate(self, entry):
        for i in Calendar.from_string(open(self.path).read()).subcomponents:
            if str(i["SUMMARY"]) == entry:
                date = str(i["CREATED"].dt).split(" ")[0]
                break
        return date

    def getNoteTags(self, entry):
        """ Workaround """
        return []
    
    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())
    
    def write(self, entry, text):
        """ Writes specified text to an entry file """
        pass

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        pass
    
    
class iPod(object):
    """ Backend for using iPod database """

    desc = 'Backend for using iPod database'
    
    ### SETTINGS DESCRIPTION ###
    path = True
    url = False
    login = False
    passwd = False

    ### FEATURE SET ###
    Tag = False
    ReadOnly = False
    URL = True

    def __init__(self, notebook):
        self.name = "iPod"
        self.notebook = notebook

        mountPoint = config.get(self.name, self.notebook)
        
        if not os.path.ismount(mountPoint):
            raise NotebookError("Specified directory is not iPod mountpoint.")
        
        self.path = os.path.join(mountPoint, "Notes")

    def getNotes(self):
        """ Returns listing of <mountpoint>/Notes """
        return os.listdir(self.path)
    
    def getDates(self):
        """ Workaround """
        return [today()]

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        return "file://" + os.path.join(self.path, entry)
        
    def getNotesByDate(self, date):
        """ Workaround """
        if date == today():
            return self.getNotes()
        else:
            return []
    
    def getNotesByTag(self, tag):
        return []
    
    def getText(self, entry):
        return open(os.path.join(self.path, entry)).read()
    
    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []
    
    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())
    
    def write(self, entry, text):
        """ Writes specified text to an entry file """
        open(os.path.join(self.path, entry), 'w').write(text)

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        os.remove(os.path.join(self.path, entry))

class Files(object):
    """ Backend for storing entries as non-organized files """

    desc = 'Backend for storing entries as non-organized files'
    ReadOnly = False
    
    ### SETTINGS DESCRIPTION ###
    path = True
    url = False
    login = False
    passwd = False

    ### FEATURE SET ###
    Tag = False
    ReadOnly = False
    URL = True

    def __init__(self, notebook):
        self.name = "Files"
        self.notebook = notebook

        self.path = config.get(self.name, self.notebook)

        self.desc = "Notes in %s directory" % (self.path)

        # Raising exception if path is wrong
        if not os.path.exists(self.path):
                raise NotebookError("Directory doesn't exist")
        
        if not os.path.isdir(self.path):
                raise NotebookError("Specified path isn't a directory")

    def getNotes(self):
        """ Returns listing of storage directory """
        return os.listdir(self.path)
    
    def getDates(self):
        """ Workaround """
        return [today()]

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        return "file://" + os.path.join(self.path, entry)

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today(): return self.getNotes()
        else: return []
    
    def getNotesByTag(self, tag):
        return []
    
    def getText(self, entry):
        return open(os.path.join(self.path, entry)).read()
    
    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []
    
    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())
    
    def write(self, entry, text):
        """ Writes specified text to an entry file """
        open(os.path.join(self.path, entry), 'w').write(text)

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        os.remove(os.path.join(self.path, entry))

class Zim(object):
    """ Backend for using Zim database """

    desc = 'Backend for using Zim database'
    ReadOnly = False

    ### SETTINGS DESCRIPTION ###
    path = True
    url = False
    login = False
    passwd = False

    ### FEATURE SET ###
    Tag = False
    ReadOnly = False
    URL = True

    def __init__(self, notebook):
        self.name = "Zim"
        self.notebook = notebook

        self.path = config.get(self.name, self.notebook)

        self.desc = "Notes in %s directory" % (self.path)

        # Raising exception if path is wrong
        if not os.path.exists(self.path):
                raise NotebookError("Directory doesn't exist")
        
        if not os.path.isdir(self.path):
                raise NotebookError("Specified path isn't a directory")

    def getNotes(self):
        t = os.walk(self.path)
        entries = []
        for st in t:
            for f in st[2]:
                if f.endswith(".txt"):
                    entries.append(os.path.splitext(f)[0].replace("_", " "))
        return entries
    
    def getPath(self, entry):
        fp = entry.replace(" ", "_") + ".txt"
        t = os.walk(self.path)
        d = {}
        for i in t:
            d[i[0]] = i[2]
        for i in d:
            if fp in d[i]:
                path = os.path.join(i, fp)
                break
        return path

    def getDates(self):
        """ Workaround """
        return [today()]

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        return "file://" + self.getPath(entry)

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today(): return self.getNotes()
        else: return []
    
    def getNotesByTag(self, tag):
        return []
    
    def getText(self, entry):
        return open(self.getPath(entry)).read()
    
    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []
    
    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())
    
    def write(self, entry, text):
        """ Writes specified text to an entry file """
        if self.noteExists(entry):
            open(self.getPath(entry), 'w').write(text)
        else:
            open(os.path.join(self.path, entry + ".txt").replace(" ", "_"), 'w').write(text)

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        os.remove(self.getPath(entry))

class DokuWiki(xmlrpclib.ServerProxy):
    """ Backend for storing entries in DokuWiki """

    desc = 'Backend for storing entries in DokuWiki'
    
    ### SETTINGS DESCRIPTION ###
    path = False
    url = True
    login = True
    passwd = True

    ### FEATURE SET ###
    Tag = False
    ReadOnly = False
    URL = True

    def __init__(self, notebook):
        self.name = "DokuWiki"
        self.notebook = notebook
        
        # Configuration
        self.url, username, pwd = config.get(self.name, self.notebook).split(";")
        xmlrpclib.ServerProxy.__init__(
            self, self.url + "/lib/exe/xmlrpc.php?" + urlencode({"u":username, "p":pwd}))
        
        self.desc = "Notes in DokuWiki at %s" % (self.url)

    def getDates(self):
        """ Workaround """
        return [today()]

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        return self.url + "/doku.php?id=%s" % (entry)

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today(): return self.getNotes()
        else: return []
    
    def getNotesByTag(self, tag):
        return []

    def getNotes(self):
        """ Walks through date directories
        and collects list of **all** entries """
        return [i["id"].encode("utf8").replace("_", " ") for i in self.wiki.getAllPages()]

    def getText(self, entry):
       """ Opens file for reading and returns text """
       return self.wiki.getPage(entry).encode("utf8")

    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []

    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return not self.getText(entry) == ""

    def write(self, entry, text):
        self.wiki.putPage(entry, text, {})

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        self.wiki.putPage(entry, "", {})

class Wixi(object):
    """ Backend for using Wixi's exported files """

    desc = "Backend for using Wixi's exported files"
    
    ### SETTINGS DESCRIPTION ###
    path = True
    url = False
    login = False
    passwd = False

    ### FEATURE SET ###
    Tag = False
    ReadOnly = True
    URL = True

    def __init__(self, notebook):
        self.name = "Wixi"
        self.notebook = notebook

        self.path = config.get(self.name, self.notebook)

        self.desc = "Notes in %s directory" % (self.path)
        
        self.export = os.path.join(self.path, "export/Wiki")

    def getNotes(self):
        """ Returns listing of storage directory """
        return [os.path.splitext(i)[0] for i in os.listdir(self.export) if i.endswith(".t2t")]
    
    def getDates(self):
        """ Workaround """
        return [today()]

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        return "file://" + os.path.join(self.export, entry + ".t2t")

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today(): return self.getNotes()
        else: return []
    
    def getNotesByTag(self, tag):
        return []
    
    def getText(self, entry):
        return open(os.path.join(self.export, entry + ".t2t")).read()
    
    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []
    
    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())
    
    def write(self, entry, text):
        """ Writes specified text to an entry file """
        open(os.path.join(self.export, entry + ".t2t"), 'w').write(text)

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        os.remove(os.path.join(self.export, entry + ".t2t"))

class RSS(object):
    """ Backend for using RSS feeds """

    desc = "Backend for using RSS feeds"
    
    ### SETTINGS DESCRIPTION ###
    path = False
    url = True
    login = False
    passwd = False

    ### FEATURE SET ###
    Tag = False
    ReadOnly = True
    URL = True

    def __init__(self, notebook):
        self.name = "RSS"
        self.notebook = notebook

        if not RSSParser:
            raise NotebookError("Python RSS FeedParser isn't installed.")
        
        self.url = config.get(self.name, self.notebook)

    def getNotes(self):
        return [i["title"].encode('utf8') for i in parse(self.url)['items']]
    
    def getDates(self):
        """ Workaround """
        return []

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        """ Returns entry's external URL """
        for i in parse(self.url)['items']:
            if i ['title'].encode('utf8') == entry:
                return i['link'].encode('utf8')

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today(): return self.getNotes()
        else: return []
    
    def getNotesByTag(self, tag):
        return []
    
    def getText(self, entry):
        for i in parse(self.url)['items']:
            if i ['title'].encode('utf8') == entry:
                return i['summary'].encode('utf8')
    
    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []
    
    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())
    
    def write(self, entry, text):
        pass

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        pass

class Mail(object):
    """ Backend for using POP3 incoming e-mail """

    desc = "Backend for using POP3 incoming e-mail"

    # Some hardcore e-mail parsing
    # http://ginstrom.com/scribbles/2007/11/19/parsing-multilingual-email-with-python

    ### SETTINGS DESCRIPTION ###
    path = False
    url = True
    login = True
    passwd = True

    ### FEATURE SET ###
    Tag = False
    ReadOnly = True
    URL = False

    def __init__(self, notebook):
        self.name = "Mail"
        self.notebook = notebook
        
        server, login, pwd = config.get(self.name, self.notebook).split(";")
       
        # Creating client
        try:
            self.client = POP3(server)
            self.client.user(login)
            self.client.pass_(pwd)
        except:
            raise NotebookError("Failed to connect to POP3 server")

    def get_charset(self, message, default="ascii"):
        """ Get the message charset """
        if message.get_content_charset():
            return message.get_content_charset()
        if message.get_charset():
            return message.get_charset()
        return default

    def get_body(self, message):
        """Get the body of the email message"""
        if message.is_multipart():
            text_parts = [part for part in typed_subpart_iterator(message,
                'text', 'plain')]
            body = []
            for part in text_parts:
                charset = self.get_charset(part, self.get_charset(message))
                body.append(unicode(part.get_payload(), charset, "replace"))
            return u"\n".join(body).strip()
        else:
            body = unicode(message.get_payload(decode=True),
                self.get_charset(message), "replace")
            return body.strip()

    def getNotes(self):
        notes = []
        for i in xrange(len(self.client.list())):
            notes.append(
                u"".join([unicode(text, charset or "ascii") for text, charset in
                decode_header(message_from_string("\n".join(self.client.retr(i + 1)[1]))["subject"])]).encode("utf8")
            )
        return notes

    def getDates(self):
        """ Workaround """
        return []

    def getTags(self):
        """ Workaround """
        return []

    def getURL(self, entry):
        pass

    def getNotesByDate(self, date):
        """ Workaround """
        if date == today(): return self.getNotes()
        else: return []

    def getNotesByTag(self, tag):
        return []

    def getText(self, entry):
        for i in xrange(len(self.client.list())):
            message = message_from_string("\n".join(self.client.retr(i + 1)[1]))
            if u"".join([unicode(text, charset or "ascii") for text, charset in
                decode_header(message["subject"])]).encode("utf8") == entry:
                return self.get_body(message).encode("utf8")

    def getNoteDate(self, entry):
        return today()

    def getNoteTags(self, entry):
        """ Workaround """
        return []

    def noteExists(self, entry):
        """ Checks if note is in all entries list """
        return (entry in self.getNotes())

    def write(self, entry, text):
        pass

    def tag(self, entry, tags):
        """ Workaround """
        pass

    def deleteNote(self, entry):
        pass

# Backends dictionary
backends = {
    "FileSystem":FileSystem,
    "Files":Files,
    "iCal":iCal,
    "iPod":iPod,
    "DokuWiki":DokuWiki,
    "RSS":RSS,
    "Mail":Mail,
    "Zim":Zim,
    "Wixi":Wixi
    }

### EXCEPTIONS ###

class NoteError(Exception):
    """ Note exception """
    pass

class NotebookError(Exception):
    """ Notebook exception """
    pass

### BASE CLASSES ###

class Notebook(object):
    """ Notebook class 
    Abstraction layer for available backends' APIs """
    def __init__(self, name):
        self.name = name
        # Selects necessary backend from dictionary above
        try:
            self.backend = backends[config.getBackend(self.name)](self.name)
        except KeyError:
            raise NotebookError("Notebook doesn't exist")
        try:
            self.markup = config.get("Markup", self.name)
        except:
            self.markup = "Wiki"
        
        self.Wiki = (self.markup == "Wiki")

    def getNotes(self):
        return self.backend.getNotes()

    def getNotesNumber(self):
        return len(self.getNotes())

    def getDates(self):
        return self.backend.getDates()

    def getDatesNumber(self):
        return len(self.getDates())

    def getTags(self):
        tags = self.backend.getTags()
        tags.sort()
        return tags

    def getTopLevelEntries(self):
        return [note for note in self.getNotes() if Note(note, self).getBacklinks() == []]

    def getTagsNumber(self):
        return len(self.getTags())

    def getNotesByDate(self, date):
        return self.backend.getNotesByDate(date)
    
    def getNotesByTag(self, tag):
        return self.backend.getNotesByTag(tag)
    
    def getNotesByText(self, text):
        return [note for note in self.getNotes() if Note(note, self).matches(text)]

    def search(self, items):
        """ Advanced search with more
        than one keyword, looks up dates, tags, titles
        and text. Google-like advanced search parameters:
            date:YYYY-M-d, year:YYYY, month:M, day:d,
            tag:str, text:str, title:str
        TODO: to implement:
            period:YYYY-M-d>YYYY-M-d, type:str """
        fullList = []
        for item in items:
            if ":" in item:
                try:
                    parameter, keyword = item.split(":")
                    if parameter == "date":
                        notes = self.getNotesByDate(keyword)
                    elif parameter == "year":
                        notes = []
                        for date in self.getDates():
                            if date.split("-")[0] == keyword:
                                notes += self.getNotesByDate(date)
                    elif parameter == "month":
                        notes = []
                        for date in self.getDates():
                            if date.split("-")[1] == keyword:
                                notes += self.getNotesByDate(date)
                    elif parameter == "day":
                        notes = []
                        for date in self.getDates():
                            if date.split("-")[2] == keyword:
                                notes += self.getNotesByDate(date)
                    elif parameter == "tag":
                        notes = self.getNotesByTag(keyword)
                    elif parameter == "text":
                        notes = self.getNotesByText(keyword)
                    elif parameter == "title":
                        notes = []
                        if Note(keyword, self).exists():
                            notes.append(keyword)
                    else:
                        notes = []
                except ValueError:
                    notes = []
            else:
                notes = self.getNotesByDate(item) + self.getNotesByTag(item) + self.getNotesByText(item)
                if Note(item, self).exists():
                    if item not in notes:
                        notes.append(item)
            fullList.append(notes)
        listIntersection = lambda list: set(list[0]) if len(list) == 1 else set(list[0]).intersection(listIntersection(list[1::]))
        return list(listIntersection(fullList))
    
    def addTag(self, tag):
        self.backend.createTag(tag)
    
    def deleteTag(self, tag):
        self.backend.removeTag(tag)

class Note(object):
    """ General Note class """
    def __init__(self, name, notebook):
        self.name = name
        
        # Entry notebook, used for backend interaction
        self.notebook = notebook

    def exists(self):
        """ Returns True if entry already exists,
        else returns False """
        return self.notebook.backend.noteExists(self.name)

    def getDate(self):
        """ Returns entry date (as str) """
        return self.notebook.backend.getNoteDate(self.name)

    def getTags(self):
        """ Returns entry tags list """
        return self.notebook.backend.getNoteTags(self.name)

    def getText(self):
        """ Returns entry text as string """
        return self.notebook.backend.getText(self.name)
    
    def matches(self, expression):
        """ Returns True if entry text matches specified
        "expression", else returns False """
        return (expression in self.getText())

    def getBacklinks(self):
        """ Returns backlinks to entry """
        return [note for note in self.notebook.getNotes()
            if Note(note, self.notebook).matches("[[%s" % (self.name))]

    def getURL(self):
        return self.notebook.backend.getURL(self.name)

    def findRelated(self):
        """ Something like backlinks, but for tags """
        notes = []
        for tag in self.getTags():
            for note in self.notebook.getNotesByTag(tag):
                if not note in notes:
                    notes.append(note)
        return notes

    def getLength(self):
        """ Returns length of "str getText()" """
        return len(self.getText())

    def write(self, text):
        """ Writes specified "text" into entry """
        self.notebook.backend.write(self.name, text)

    def tag(self, tags):
        """ Tags entry with specified "tags".
        Tag is something like category, but unlike
        it entry can be placed into more than one tag """
        self.notebook.backend.tag(self.name, tags)

    def copyToNotebook(self, notebooks, move=False):
        """ Copies entry from one notebook to
        all notebooks from notebooks list (this is backend-universal).
        Use move=True to move entry instead of copying"""
        for nb in notebooks:
            Note(self.name, Notebook(nb)).write(self.getText())
        if move:
            self.delete()
         
    def import_(self, file):
        """ Imports locally available text
        file into notebook as entry """
        self.write(open(file).read())

    def delete(self):
        """ Deletes entry """
        self.notebook.backend.deleteNote(self.name)

    def rename(self, name):
        """ Renames entry """
        note = Note(name, self.notebook)
        note.write(self.getText())
        note.tag(self.getTags())
        self.delete()
