#!/usr/bin/env python
# -*- coding: utf-8 -*
#
#       Copyright (C) 2007-2008 Peter Hoffmann <tosh54@gmail.com>
#       Copyright (C) 2007-2008 Simonenko Sergey <gforgx@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor,
#       Boston, MA  02110-1301  USA

"""
pycreole is a python implementation of the creole wiki standard."""

__author__  = "Peter Hoffmann <tosh54@gmail.com>"
__version__ = "0.25"
__date__    = "2008-09-14"

import sys
import re

from math import *

try:
    import lxml.etree as ET
except ImportError:
    try:
        import xml.etree.ElementTree as ET # python >= 2.5
    except ImportError:
        import elementtree.ElementTree as ET #effbots pure python module

#import functools

# see http://effbot.org/zone/element-builder.htm
class _E(object):
    def __call__(self, tag, *children, **attrib):
        elem = ET.Element(tag, attrib)
        for item in children:
            if isinstance(item, dict):
                #elem.attrib.update(item)
                #throws exception with lxml so do it manually
                for k in item:
                    elem.attrib[k] = item[k]
            elif isinstance(item, basestring):
                if len(elem):
                    elem[-1].tail = (elem[-1].tail or "") + item
                else:
                    elem.text = (elem.text or "") + item
            elif ET.iselement(item):
                elem.append(item)
            else:
                raise TypeError("bad argument: %r" % item)
        return elem

    #def __getattr__(self, tag):
    #    return functools.partial(self, tag)

# create factory object
E = _E()


def strong(t):
    return E("strong", *markup2et(t))

def strike(t):
    return E("strike", *markup2et(t))

def em(t):
    return E("em", *markup2et(t))

def underline(t):
    return E("u", *markup2et(t))

def monospace(t):
    return E("tt", *markup2et(t))

def highlight(t):
    return E("span", {"style": "background: yellow;"}, *markup2et(t))

def linebreak(t):
    return E("br")

def image(t):
    try:
        src, alt = t.split("|",1)
    except ValueError:
        src = alt = t
    return E("img",{"src":src,"alt":alt})

def link(t):
    return E("a", t, {"href":t})

def wikilink(t):
    try:
        link, text = t.split("|",1)
        link = link.strip()
    except ValueError:
        link = text = t
    return E("a", text, {"href":link})

def camelcase(t):
    return E("a", t, {"href":t})

def calculation(t):
    try:
        result = eval(t)
    except Exception, err:
        result = err
    return E("tt", str(result))

def heading(t):
    level = 0
    while t[level] == "=":
        level+=1
    t = t[level:].strip()
    return E("h"+str(level),t) 

def paragraph(t):
    """split by blank lines and apply markup"""
    #TODO does this work on windows?
    parts = re.split(re.compile(r"^\s*\n",re.M),t)
    result =[]
    for part in parts:
        if not re.match("^\s*$",part):
            result.append(E("p", *markup2et(part)))
    return result

def pre(t):
	return E("pre", t)

def hr(t):
    return E("hr")

def prepare_list(s):
    # Splittet die Zeilen auf und fuegt mehrzeilige listenelemente (fangen nicht mit * an) zusammen
    # ausserdem weden die anzahl der fuehrenden * gezaehlt 
    # gibt tupel (anzfuehr*, lineohne*) zurueck
    lines = []
    for line in s.splitlines():
        if not line.startswith("*"):
            lines[-1][1] += line
            continue
        level = 0
        while line[level] == "*":
            level += 1
        lines.append([level, line[level:].lstrip()]) 
    
    #fix lists which do not starting with first level
    #TODO test this  
    if len(lines)>0 and lines[0][0] >1:
        lines.insert(0,[1,""])
    return lines

def genlist(lines):
    hold = []
    curline = None
    parts = []
    for line in lines:
        if line[0] > 1:
            hold.append(line)
            hold[-1][0] -= 1
        else:
            if curline is None:
                curline = line
                continue
            if len(hold) >0:
                ct = markup2et(curline[1])
                ct.append(genlist(hold))
                parts.append(E("li", *ct))
                hold = []
            else:
                parts.append(E("li", *markup2et(curline[1])))   
            curline = line
    
    if curline is None:
        curline = [0, ""]
    if len(hold) >0:
        ct = markup2et(curline[1])
        ct.append(genlist(hold))
        parts.append(E("li", *ct))
    else:
        parts.append(E("li", *markup2et(curline[1])))

    return E("ul", *parts)

def list(t):
    return genlist(prepare_list(t))

def split_row(text):
    """split row by | and watch for link ambiguity"""
    #TODO what about escaped characters
    #TODO this is a hack because of the tables and link ambiguity
    link_def = re.compile("(?:{{(?P<image>.*?)}})|(?:\[\[(?P<wikilink>.*?)\]\])")
    pos = 0
    result = [u""]
    for match in link_def.finditer(text):
        p = text[pos:match.start()].split("|")
        result[-1]+= p[0]
        for a in p[1:]:
            result.append(a)
        result[-1] += text[match.start():match.end()]
        pos = match.end()
    p = text[pos:].split("|")
    result[-1]+= p[0]
    for a in p[1:]:
        result.append(a)
    return result

def table(t):
    result = []
    for (linenr, line) in enumerate(t.splitlines()):
        line = line[1:].strip()
        if line.endswith("|"):
            line = line[:-1]
        parts = split_row(line)
        if linenr == 0 and parts[0].strip().startswith("="): #do we have table headings
            newparts = []
            for p in parts:
                if p.strip().startswith("="):
                    newparts.append(p.strip()[1:])
                else:
                    newparts.append(p)
            th_list = []
            for th in newparts:
                th_list.append(E("th", *markup2et(th)))
            result.append(E("tr", *th_list))
        else:
            td_list=[]
            for td in parts:
                td_list.append(E("td", *markup2et(td)))
            
            result.append(E("tr", *td_list))

    #TODO remove rules, border??
    return E("table", {"border":"0.3"}, *result)

markup_def = re.compile(
        r"(?:\*\*(?P<strong>.*?)\*\*)|"
        + r"(?:__(?P<underline>.*?)__)|"
        + r"(?:--(?P<strike>.*?)--)|"
        + r"(?:{{(?P<image>.*?)}})|"
        + r"(?://(?P<em>.*?[^:])//)|"
        + r"(?:%%(?P<highlight>.*?)%%)|"
        + r"(?:##(?P<monospace>.*?)##)|"
        + r"(?P<linebreak>\\\\\n)|"
        + r"(?P<link>(?:http|https|ftp|file)://\S*)|"
        + r"(?:\[\[(?P<wikilink>.*?)\]\])|"
        + r"(?P<camelcase>(?:[A-Z]+[a-z]+[A-Z])+\S*)|"
        + r"(?:\~(?P<calculation>.*?)\~)",re.MULTILINE|re.DOTALL)

block_def = re.compile(
        r"(?:^\{\{\{(?P<pre>.*?)^\}\}\} *$)|"
        + r"(?:(?P<heading>^=+ .*?)=* *$)|"
        + r"(?P<list>(?:(?:^\*+ .*?$\n?)(?:[ \t]+.*?$\n?)*)+)|"
        + r"(?P<table>(?:^\|.*?\n)+)|"
        + r"(?P<hr>^---- *$)",re.MULTILINE|re.DOTALL)


def applyPattern(pattern, text, func=None):
    """apply makup to a text block and handle non machting text by function func"""
    pos = 0
    result = []
    for match in pattern.finditer(text):
        ntext = text[pos:match.start()]
        #handle text between matches
        if len(ntext)>0:
            if func is not None:
                #kommt nur im paragraph Fall vor
                result += func(ntext)
            else:
                result.append(ntext)
        pos = match.end()
        for type, hit in match.groupdict().items():
            if hit:
                #call function with name of match
                result.append(globals()[type](hit))

    ntext = text[pos:]
    #handle text between last match and end
    if len(ntext)>0:
        if func is not None:
            #kommt nur im paragraph fall vor
            result += func(ntext)
        else:   
            result.append(ntext)
    return result

def markup2et(text):
    """returns a list of text or/and Element elements"""
    return applyPattern(markup_def, text)

def wiki2et(text):
    if not text.endswith("\n"):
        text += "\n"
    return E("div", *applyPattern(block_def, text, paragraph))

def wiki2html(text, encoding="ascii"):
    """convert wiki markup to html"""
    if ET.__name__ == 'lxml.etree':
        return ET.tostring(wiki2et(text), encoding=encoding, pretty_print=True)  
    else:
        return ET.tostring(wiki2et(text), encoding=encoding)  

def main(argv=None):
    if argv is None:
        argv = sys.argv
            
    text = sys.stdin.read()
    #TODO fix encoding
    text = text.decode("utf-8")

    #if sys.stdout.encoding is None:
        #if python cannot get an encoding set it to utf-8
        #e.g useful for shell redirection
        #see http://wiki.python.org/moin/ShellRedirectionFails
     #   import codecs
     #   sys.stdout = codecs.getwriter("utf-8")(sys.__stdout__)

    #TODO fix encoding
    html = wiki2html(text, "utf-8")
    #print type(html)
    #sys.stdout.write()
    print html   
if __name__ == "__main__":
    main()
    
